/*
 * GTK Whistler Theme
 *
 * Copyright (C) 2000 Kazuki Yasumatsu <kyasu@asahi-net.email.ne.jp>
 */

#include <math.h>
#include <gtk/gtk.h>

#define SCROLLBAR_WORKAROUND

#define DETAIL(xx)	((detail) && (!strcmp(xx, detail)))

#define whistler_light_gc	(style->base_gc[GTK_STATE_PRELIGHT])
#define whistler_dark_gc	(style->base_gc[GTK_STATE_INSENSITIVE])

extern GtkStyleClass whistler_default_class;

/* internal functions */
static void         draw_hline(GtkStyle * style,
			       GdkWindow * window,
			       GtkStateType state_type,
			       GdkRectangle * area,
			       GtkWidget * widget,
			       gchar * detail,
			       gint x1,
			       gint x2,
			       gint y);
static void         draw_vline(GtkStyle * style,
			       GdkWindow * window,
			       GtkStateType state_type,
			       GdkRectangle * area,
			       GtkWidget * widget,
			       gchar * detail,
			       gint y1,
			       gint y2,
			       gint x);
static void         draw_shadow(GtkStyle * style,
				GdkWindow * window,
				GtkStateType state_type,
				GtkShadowType shadow_type,
				GdkRectangle * area,
				GtkWidget * widget,
				gchar * detail,
				gint x,
				gint y,
				gint width,
				gint height);

static void         draw_polygon(GtkStyle * style,
				 GdkWindow * window,
				 GtkStateType state_type,
				 GtkShadowType shadow_type,
				 GdkRectangle * area,
				 GtkWidget * widget,
				 gchar * detail,
				 GdkPoint * point,
				 gint npoints,
				 gint fill);
static void         draw_arrow(GtkStyle * style,
			       GdkWindow * window,
			       GtkStateType state_type,
			       GtkShadowType shadow_type,
			       GdkRectangle * area,
			       GtkWidget * widget,
			       gchar * detail,
			       GtkArrowType arrow_type,
			       gint fill,
			       gint x,
			       gint y,
			       gint width,
			       gint height);
static void         draw_diamond(GtkStyle * style,
				 GdkWindow * window,
				 GtkStateType state_type,
				 GtkShadowType shadow_type,
				 GdkRectangle * area,
				 GtkWidget * widget,
				 gchar * detail,
				 gint x,
				 gint y,
				 gint width,
				 gint height);
static void         draw_oval(GtkStyle * style,
			      GdkWindow * window,
			      GtkStateType state_type,
			      GtkShadowType shadow_type,
			      GdkRectangle * area,
			      GtkWidget * widget,
			      gchar * detail,
			      gint x,
			      gint y,
			      gint width,
			      gint height);
static void         draw_string(GtkStyle * style,
				GdkWindow * window,
				GtkStateType state_type,
				GdkRectangle * area,
				GtkWidget * widget,
				gchar * detail,
				gint x,
				gint y,
				const gchar * string);
static void         draw_box(GtkStyle * style,
			     GdkWindow * window,
			     GtkStateType state_type,
			     GtkShadowType shadow_type,
			     GdkRectangle * area,
			     GtkWidget * widget,
			     gchar * detail,
			     gint x,
			     gint y,
			     gint width,
			     gint height);
static void         draw_flat_box(GtkStyle * style,
				  GdkWindow * window,
				  GtkStateType state_type,
				  GtkShadowType shadow_type,
				  GdkRectangle * area,
				  GtkWidget * widget,
				  gchar * detail,
				  gint x,
				  gint y,
				  gint width,
				  gint height);
static void         draw_check(GtkStyle * style,
			       GdkWindow * window,
			       GtkStateType state_type,
			       GtkShadowType shadow_type,
			       GdkRectangle * area,
			       GtkWidget * widget,
			       gchar * detail,
			       gint x,
			       gint y,
			       gint width,
			       gint height);
static void         draw_option(GtkStyle * style,
				GdkWindow * window,
				GtkStateType state_type,
				GtkShadowType shadow_type,
				GdkRectangle * area,
				GtkWidget * widget,
				gchar * detail,
				gint x,
				gint y,
				gint width,
				gint height);
static void         draw_cross(GtkStyle * style,
			       GdkWindow * window,
			       GtkStateType state_type,
			       GtkShadowType shadow_type,
			       GdkRectangle * area,
			       GtkWidget * widget,
			       gchar * detail,
			       gint x,
			       gint y,
			       gint width,
			       gint height);
static void         draw_ramp(GtkStyle * style,
			      GdkWindow * window,
			      GtkStateType state_type,
			      GtkShadowType shadow_type,
			      GdkRectangle * area,
			      GtkWidget * widget,
			      gchar * detail,
			      GtkArrowType arrow_type,
			      gint x,
			      gint y,
			      gint width,
			      gint height);
static void         draw_tab(GtkStyle * style,
			     GdkWindow * window,
			     GtkStateType state_type,
			     GtkShadowType shadow_type,
			     GdkRectangle * area,
			     GtkWidget * widget,
			     gchar * detail,
			     gint x,
			     gint y,
			     gint width,
			     gint height);
static void          draw_shadow_gap(GtkStyle * style,
				    GdkWindow * window,
				    GtkStateType state_type,
				    GtkShadowType shadow_type,
				    GdkRectangle * area,
				    GtkWidget * widget,
				    gchar * detail,
				    gint x,
				    gint y,
				    gint width,
				    gint height,
				    GtkPositionType gap_side,
				    gint gap_x,
				    gint gap_width);
static void         draw_box_gap(GtkStyle * style,
				 GdkWindow * window,
				 GtkStateType state_type,
				 GtkShadowType shadow_type,
				 GdkRectangle * area,
				 GtkWidget * widget,
				 gchar * detail,
				 gint x,
				 gint y,
				 gint width,
				 gint height,
				 GtkPositionType gap_side,
				 gint gap_x,
				 gint gap_width);
static void         draw_extension(GtkStyle * style,
				   GdkWindow * window,
				   GtkStateType state_type,
				   GtkShadowType shadow_type,
				   GdkRectangle * area,
				   GtkWidget * widget,
				   gchar * detail,
				   gint x,
				   gint y,
				   gint width,
				   gint height,
				   GtkPositionType gap_side);
static void         draw_focus(GtkStyle * style,
			       GdkWindow * window,
			       GdkRectangle * area,
			       GtkWidget * widget,
			       gchar * detail,
			       gint x,
			       gint y,
			       gint width,
			       gint height);
static void         draw_slider(GtkStyle * style,
				GdkWindow * window,
				GtkStateType state_type,
				GtkShadowType shadow_type,
				GdkRectangle * area,
				GtkWidget * widget,
				gchar * detail,
				gint x,
				gint y,
				gint width,
				gint height,
				GtkOrientation orientation);
static void         draw_handle(GtkStyle * style,
				GdkWindow * window,
				GtkStateType state_type,
				GtkShadowType shadow_type,
				GdkRectangle * area,
				GtkWidget * widget,
				gchar * detail,
				gint x,
				gint y,
				gint width,
				gint height,
				GtkOrientation orientation);

/* internal data structs */

GtkStyleClass whistler_default_class =
{
    1,
    1,
    draw_hline,
    draw_vline,
    draw_shadow,
    draw_polygon,
    draw_arrow,
    draw_diamond,
    draw_oval,
    draw_string,
    draw_box,
    draw_flat_box,
    draw_check,
    draw_option,
    draw_cross,
    draw_ramp,
    draw_tab,
    draw_shadow_gap,
    draw_box_gap,
    draw_extension,
    draw_focus,
    draw_slider,
    draw_handle
};

static void
draw_hline(GtkStyle * style,
	   GdkWindow * window,
	   GtkStateType state_type,
	   GdkRectangle * area,
	   GtkWidget * widget,
	   gchar * detail,
	   gint x1,
	   gint x2,
	   gint y)
{
    GdkGC *gc;

    g_return_if_fail(style != NULL);
    g_return_if_fail(window != NULL);

    gc = whistler_dark_gc;

    if (area) gdk_gc_set_clip_rectangle(gc, area);

    gdk_draw_line(window, gc, x1, y, x2, y);

    if (area) gdk_gc_set_clip_rectangle(gc, NULL);
}

static void
draw_vline(GtkStyle * style,
	   GdkWindow * window,
	   GtkStateType state_type,
	   GdkRectangle * area,
	   GtkWidget * widget,
	   gchar * detail,
	   gint y1,
	   gint y2,
	   gint x)
{
    GdkGC *gc;

    g_return_if_fail(style != NULL);
    g_return_if_fail(window != NULL);

    gc = whistler_dark_gc;

    if (area) gdk_gc_set_clip_rectangle(gc, area);

    gdk_draw_line(window, gc, x, y1, x, y2);

    if (area) gdk_gc_set_clip_rectangle(gc, NULL);
}

static void
draw_shadow(GtkStyle * style,
	    GdkWindow * window,
	    GtkStateType state_type,
	    GtkShadowType shadow_type,
	    GdkRectangle * area,
	    GtkWidget * widget,
	    gchar * detail,
	    gint x,
	    gint y,
	    gint width,
	    gint height)
{
    GdkGC *darkgc, *lightgc, *whitgc;

    g_return_if_fail(style != NULL);
    g_return_if_fail(window != NULL);

    switch (shadow_type) {
    case GTK_SHADOW_IN:
    case GTK_SHADOW_OUT:
    case GTK_SHADOW_ETCHED_IN:
    case GTK_SHADOW_ETCHED_OUT:
	break;
    default:
	return;
    }

    if (width == -1 && height == -1)
	gdk_window_get_size(window, &width, &height);
    else if (width == -1)
	gdk_window_get_size(window, &width, NULL);
    else if (height == -1)
	gdk_window_get_size(window, NULL, &height);

    darkgc  = whistler_dark_gc;
    lightgc = whistler_light_gc;
    whitgc  = style->white_gc;

    if (area) {
	gdk_gc_set_clip_rectangle(darkgc,  area);
	gdk_gc_set_clip_rectangle(lightgc, area);
	gdk_gc_set_clip_rectangle(whitgc,  area);
    }

    gdk_draw_line(window, darkgc, x, y+height-1, x+width-1, y+height-1);
    gdk_draw_line(window, darkgc, x+width-1, y, x+width-1, y+height-1);
    gdk_draw_line(window, darkgc, x, y, x+width-1, y);
    gdk_draw_line(window, darkgc, x, y, x, y+height-1);

    if (DETAIL("button") || DETAIL("buttondefault") ||
	DETAIL("togglebutton") || DETAIL("optionmenu"))
    {
	switch (shadow_type) {
	case GTK_SHADOW_IN:
	case GTK_SHADOW_ETCHED_IN:
	    gdk_draw_line(window, lightgc, x+1, y+1, x+width-2, y+1);
	    gdk_draw_line(window, lightgc, x+1, y+1, x+1, y+height-2);
	    gdk_draw_line(window, whitgc, x+1, y+height-2, x+width-2, y+height-2);
	    gdk_draw_line(window, whitgc, x+width-2, y+1, x+width-2, y+height-2);
	    break;

	case GTK_SHADOW_OUT:
	case GTK_SHADOW_ETCHED_OUT:
	    gdk_draw_line(window, lightgc, x+1, y+height-2, x+width-2, y+height-2);
	    gdk_draw_line(window, lightgc, x+width-2, y+1, x+width-2, y+height-2);
	    gdk_draw_line(window, whitgc, x+1, y+1, x+width-2, y+1);
	    gdk_draw_line(window, whitgc, x+1, y+1, x+1, y+height-2);
	    break;
	}
    }

    if (area) {
	gdk_gc_set_clip_rectangle(darkgc,  NULL);
	gdk_gc_set_clip_rectangle(lightgc, NULL);
	gdk_gc_set_clip_rectangle(whitgc,  NULL);
    }
}

static void
draw_polygon(GtkStyle * style,
	     GdkWindow * window,
	     GtkStateType state_type,
	     GtkShadowType shadow_type,
	     GdkRectangle * area,
	     GtkWidget * widget,
	     gchar * detail,
	     GdkPoint * points,
	     gint npoints,
	     gint fill)
{
    GdkGC *fgc, *bgc;

    g_return_if_fail(style != NULL);
    g_return_if_fail(window != NULL);
    g_return_if_fail(points != NULL);

    fgc = whistler_dark_gc;
    bgc = style->bg_gc[state_type];

    if (area) {
	gdk_gc_set_clip_rectangle(fgc, area);
	gdk_gc_set_clip_rectangle(bgc, area);
    }

    if (fill)
	gdk_draw_polygon(window, bgc, TRUE, points, npoints);

    gdk_draw_polygon(window, fgc, FALSE, points, npoints);

    if (area) {
	gdk_gc_set_clip_rectangle(fgc, NULL);
	gdk_gc_set_clip_rectangle(bgc, NULL);
    }
}

static void
draw_arrow(GtkStyle     *style,
	   GdkWindow    *window,
	   GtkStateType  state_type,
	   GtkShadowType shadow_type,
	   GdkRectangle *area,
	   GtkWidget    *widget,
	   gchar        *detail,
	   GtkArrowType  arrow_type,
	   gboolean      fill,
	   gint          x,
	   gint          y,
	   gint          width,
	   gint          height)
{
    GdkGC *gc;
    gboolean border;
    gint xthik, ythik;
    gint half_width, half_height;
    GdkPoint points[3];

    g_return_if_fail(style != NULL);
    g_return_if_fail(window != NULL);

    if (width == -1 && height == -1)
	gdk_window_get_size(window, &width, &height);
    else if (width == -1)
	gdk_window_get_size(window, &width, NULL);
    else if (height == -1)
	gdk_window_get_size(window, NULL, &height);

    if (state_type == GTK_STATE_INSENSITIVE)
	gc = style->fg_gc[state_type];
    else
	gc = style->black_gc;

    border = FALSE;
    if (DETAIL("spinbutton") || GTK_IS_NOTEBOOK(widget)) {
	border = TRUE;
    }
    else if (GTK_IS_SCROLLBAR(widget)) {
#if 0
	border = TRUE;
#endif
	if (state_type == GTK_STATE_ACTIVE)
	    state_type = GTK_STATE_PRELIGHT;
#ifdef SCROLLBAR_WORKAROUND
	{
	gint tw, th;
	gdk_window_get_size(GTK_RANGE(widget)->trough, &tw, &th);
	if (GTK_IS_VSCROLLBAR(widget)) {
	    if (width != tw - 2) {
		width = tw - 2;
		gdk_window_resize(window, width, height);
	    }
	}
	else {
	    if (height != th - 2) {
		height = th - 2;
		gdk_window_resize(window, width, height);
	    }
	}
	}
#endif
    }

    if (fill) {
	if (border)
	    gtk_paint_box(style, window, state_type, shadow_type, area,
			  widget, detail, x, y, width, height);
	else
	    gtk_paint_box(style, window, state_type, GTK_SHADOW_NONE, area,
			  widget, detail, x, y, width, height);
    }

#if 0
    xthik = style->klass->xthickness;
    ythik = style->klass->ythickness;
#else
    xthik = 1;
    ythik = 1;
#endif

    if (border) {
	x += (xthik + 1);
	y += (ythik + 1);
	width -= (xthik + 1) * 2;
	height -= (ythik + 1) * 2;
    }
    else {
	x += 1;
	y += 1;
	width -= 2;
	height -= 2;
    }

    if (!(width & 1))
	width--;
    if (!(height & 1))
	height--;

    half_width = width / 2;
    half_height = height / 2;

    switch (arrow_type) {
    case GTK_ARROW_UP:
	points[0].x = x;
	points[0].y = y + half_height + (half_width / 2);
	points[1].x = x + width - 1;
	points[1].y = y + half_height + (half_width / 2);
	points[2].x = x + half_width;
	points[2].y = y + half_height - (half_width / 2);
	break;

    case GTK_ARROW_DOWN:
	points[0].x = x;
	points[0].y = y + half_height - (half_width / 2);
	points[1].x = x + width - 1;
	points[1].y = y + half_height - (half_width / 2);
	points[2].x = x + half_width;
	points[2].y = y + half_height + (half_width / 2);
	break;

    case GTK_ARROW_LEFT:
	points[0].x = x + half_width + (half_height / 2);
	points[0].y = y;
	points[1].x = x + half_width + (half_height / 2);
	points[1].y = y + height - 1;
	points[2].x = x + half_width - (half_height / 2);
	points[2].y = y + half_height;
	break;

    case GTK_ARROW_RIGHT:
	points[0].x = x + half_width - (half_height / 2);
	points[0].y = y;
	points[1].x = x + half_width - (half_height / 2);
	points[1].y = y + height - 1;
	points[2].x = x + half_width + (half_height / 2);
	points[2].y = y + half_height;
	break;
    }

    if (area) gdk_gc_set_clip_rectangle(gc, area);

    gdk_draw_polygon(window, gc, TRUE, points, 3);
    gdk_draw_polygon(window, gc, FALSE, points, 3);

    if (area) gdk_gc_set_clip_rectangle(gc, NULL);
}

static void
draw_diamond(GtkStyle * style,
	     GdkWindow * window,
	     GtkStateType state_type,
	     GtkShadowType shadow_type,
	     GdkRectangle * area,
	     GtkWidget * widget,
	     gchar * detail,
	     gint x,
	     gint y,
	     gint width,
	     gint height)
{
    GdkGC *fgc, *bgc, *dgc;
    gint half_width, half_height;
    GdkPoint points[4];

    g_return_if_fail(style != NULL);
    g_return_if_fail(window != NULL);

    if (width == -1 && height == -1)
	gdk_window_get_size(window, &width, &height);
    else if (width == -1)
	gdk_window_get_size(window, &width, NULL);
    else if (height == -1)
	gdk_window_get_size(window, NULL, &height);

    x -= 1;
    y -= 1;
    width += 2;
    height += 2;

    if ((width & 1))
	width--;
    if ((height & 1))
	height--;

    if (state_type == GTK_STATE_INSENSITIVE) {
	fgc = style->fg_gc[state_type];
	bgc = style->bg_gc[GTK_STATE_NORMAL];
    }
    else {
	fgc = style->black_gc;
	bgc = style->white_gc;
    }
    dgc  = whistler_dark_gc;

    if (area) {
	gdk_gc_set_clip_rectangle(fgc, area);
	gdk_gc_set_clip_rectangle(bgc, area);
	gdk_gc_set_clip_rectangle(dgc, area);
    }

    half_width = width / 2;
    half_height = height / 2;

    points[0].x = x + half_width;
    points[0].y = y;
    points[1].x = x + width;
    points[1].y = y + half_height;
    points[2].x = x + half_width;
    points[2].y = y + height;
    points[3].x = x;
    points[3].y = y + half_height;

    gdk_draw_polygon(window, bgc, TRUE, points, 4);
    gdk_draw_polygon(window, dgc, FALSE, points, 4);

    if (shadow_type == GTK_SHADOW_IN) {
	points[0].y += 4;
	points[1].x -= 4;
	points[2].y -= 4;
	points[3].x += 4;

	gdk_draw_polygon(window, fgc, TRUE, points, 4);
	gdk_draw_polygon(window, fgc, FALSE, points, 4);
    }

    if (area) {
	gdk_gc_set_clip_rectangle(fgc, NULL);
	gdk_gc_set_clip_rectangle(bgc, NULL);
	gdk_gc_set_clip_rectangle(dgc, NULL);
    }
}

static void
draw_oval(GtkStyle * style,
	  GdkWindow * window,
	  GtkStateType state_type,
	  GtkShadowType shadow_type,
	  GdkRectangle * area,
	  GtkWidget * widget,
	  gchar * detail,
	  gint x,
	  gint y,
	  gint width,
	  gint height)
{
    g_return_if_fail(style != NULL);
    g_return_if_fail(window != NULL);
}

static void
draw_string(GtkStyle * style,
	    GdkWindow * window,
	    GtkStateType state_type,
	    GdkRectangle * area,
	    GtkWidget * widget,
	    gchar * detail,
	    gint x,
	    gint y,
	    const gchar * string)
{
    GdkGC *gc;

    g_return_if_fail(style != NULL);
    g_return_if_fail(window != NULL);

    gc = style->fg_gc[state_type];

    if (area) gdk_gc_set_clip_rectangle(gc, area);

    gdk_draw_string(window, style->font, gc, x, y, string);

    if (area) gdk_gc_set_clip_rectangle(gc, NULL);
}

static void
draw_box(GtkStyle * style,
	 GdkWindow * window,
	 GtkStateType state_type,
	 GtkShadowType shadow_type,
	 GdkRectangle * area,
	 GtkWidget * widget,
	 gchar * detail,
	 gint x,
	 gint y,
	 gint width,
	 gint height)
{
    gint xoff, yoff;
    gboolean shadow;
    GdkGC *gc;

    g_return_if_fail(style != NULL);
    g_return_if_fail(window != NULL);

    if (width == -1 && height == -1)
	gdk_window_get_size(window, &width, &height);
    else if (width == -1)
	gdk_window_get_size(window, &width, NULL);
    else if (height == -1)
	gdk_window_get_size(window, NULL, &height);

    xoff = yoff = 0;
    shadow = TRUE;
    gc = NULL;

    if (DETAIL("menuitem") || DETAIL("menubar")) {
	shadow = FALSE;
    }
    else if (DETAIL("trough")) {
	if (GTK_IS_PROGRESS_BAR(widget))
	    gc = style->base_gc[GTK_STATE_NORMAL];
	else
	    gc = style->bg_gc[GTK_STATE_NORMAL];
    }
    else if (DETAIL("bar")) {
	xoff = yoff = 1;
	shadow = FALSE;
	gc = style->bg_gc[GTK_STATE_SELECTED];
    }
    else if (DETAIL("slider")) {
	if (GTK_IS_SCROLLBAR(widget)) {
	    if (state_type == GTK_STATE_INSENSITIVE)
		gc = style->bg_gc[GTK_STATE_NORMAL];
	    else if (state_type != GTK_STATE_PRELIGHT)
		gc = style->base_gc[GTK_STATE_NORMAL];
#ifdef SCROLLBAR_WORKAROUND
	    {
	    gint tw, th;
	    gdk_window_get_size(GTK_RANGE(widget)->trough, &tw, &th);
	    if (GTK_IS_VSCROLLBAR(widget)) {
		if (width != tw - 2) {
		    width = tw - 2;
		    gdk_window_resize(window, width, height);
		}
	    }
	    else {
		if (height != th - 2) {
		    height = th - 2;
		    gdk_window_resize(window, width, height);
		}
	    }
	    }
#endif
	}
    }

    if (!gc ||
	!style->bg_pixmap[state_type] ||
	gdk_window_get_type(window) == GDK_WINDOW_PIXMAP)
    {
	if (gc == NULL) gc = style->bg_gc[state_type];

	if (area) gdk_gc_set_clip_rectangle(gc, area);

	gdk_draw_rectangle(window, gc, TRUE,
			   x + xoff, y + yoff,
			   width - (xoff * 2), height - (yoff * 2));

	if (area) gdk_gc_set_clip_rectangle(gc, NULL);
    }
    else {
	gtk_style_apply_default_background(style, window,
					   widget && !GTK_WIDGET_NO_WINDOW(widget),
					   state_type, area,
					   x + xoff, y + yoff,
					   width - (xoff * 2),
					   height - (yoff * 2));
    }

    if (shadow)
	gtk_paint_shadow(style, window, state_type, shadow_type, area, widget,
			 detail, x, y, width, height);
}

static void
draw_flat_box(GtkStyle * style,
	      GdkWindow * window,
	      GtkStateType state_type,
	      GtkShadowType shadow_type,
	      GdkRectangle * area,
	      GtkWidget * widget,
	      gchar * detail,
	      gint x,
	      gint y,
	      gint width,
	      gint height)
{
    GdkGC *gc;

    g_return_if_fail(style != NULL);
    g_return_if_fail(window != NULL);

    if (width == -1 && height == -1)
	gdk_window_get_size(window, &width, &height);
    else if (width == -1)
	gdk_window_get_size(window, &width, NULL);
    else if (height == -1)
	gdk_window_get_size(window, NULL, &height);

    gc = style->bg_gc[state_type];
    if (DETAIL("selected"))
	gc = style->bg_gc[GTK_STATE_SELECTED];
    else if (DETAIL("text") && state_type == GTK_STATE_SELECTED)
	gc = style->bg_gc[GTK_STATE_SELECTED];
    else if (DETAIL("viewportbin"))
	gc = style->bg_gc[GTK_STATE_NORMAL];
    else if (DETAIL("entry_bg")) {
	if (state_type == GTK_STATE_INSENSITIVE)
	    gc = style->bg_gc[GTK_STATE_NORMAL];
	else
	    gc = style->base_gc[GTK_STATE_NORMAL];
    }

    if (gc != style->bg_gc[state_type] ||
	!style->bg_pixmap[state_type] ||
	gdk_window_get_type(window) == GDK_WINDOW_PIXMAP)
    {
	if (area) gdk_gc_set_clip_rectangle(gc, area);

	gdk_draw_rectangle(window, gc, TRUE, x, y, width, height);

	if (area) gdk_gc_set_clip_rectangle(gc, NULL);

	if (DETAIL("tooltip")) {
	    gc = whistler_dark_gc;

	    if (area) gdk_gc_set_clip_rectangle(gc, area);
	    gdk_draw_rectangle(window, gc, FALSE, x, y, width - 1, height - 1);
	    if (area) gdk_gc_set_clip_rectangle(gc, NULL);
	}
    }
    else {
	gtk_style_apply_default_background(style, window,
					   widget && !GTK_WIDGET_NO_WINDOW(widget),
					   state_type, area,
					   x, y, width, height);
    }
}

static void
draw_check(GtkStyle * style,
	   GdkWindow * window,
	   GtkStateType state_type,
	   GtkShadowType shadow_type,
	   GdkRectangle * area,
	   GtkWidget * widget,
	   gchar * detail,
	   gint x,
	   gint y,
	   gint width,
	   gint height)
{
    GdkGC *fgc, *bgc;
    gint xx, yy, ww, hh;
    gint xthick, ythick;

    g_return_if_fail (style != NULL);
    g_return_if_fail (window != NULL);

    if (width == -1 && height == -1)
	gdk_window_get_size(window, &width, &height);
    else if (width == -1)
	gdk_window_get_size(window, &width, NULL);
    else if (height == -1)
	gdk_window_get_size(window, NULL, &height);

    if (state_type == GTK_STATE_INSENSITIVE) {
	fgc = style->fg_gc[state_type];
	bgc = style->bg_gc[GTK_STATE_NORMAL];
    }
    else {
	fgc = style->black_gc;
	bgc = style->white_gc;
    }

    if (area) {
	gdk_gc_set_clip_rectangle(fgc, area);
	gdk_gc_set_clip_rectangle(bgc, area);
    }

    gdk_draw_rectangle(window, bgc, TRUE, x, y, width, height);

    if (shadow_type == GTK_SHADOW_IN) {
	xx = x + 1;
	yy = y + 1;
	ww = width - 2;
	hh = height - 2;

	gdk_draw_line(window, fgc,
		      xx + ww - 1, yy,
		      xx + (ww / 3), yy + hh - 3);
	gdk_draw_line(window, fgc,
		      xx, yy + (2 * hh / 3) - 2,
		      xx + (ww / 3), yy + hh - 3);
	gdk_draw_line(window, fgc,
		      xx + ww - 1, yy + 1,
		      xx + (ww / 3), yy + hh - 2);
	gdk_draw_line(window, fgc,
		      xx, yy + (2 * hh / 3) - 1,
		      xx + (ww / 3), yy + hh - 2);
	gdk_draw_line(window, fgc,
		      xx + ww - 1, yy + 2,
		      xx + (ww / 3), yy + hh - 1);
	gdk_draw_line(window, fgc,
		      xx, yy + (2 * hh / 3),
		      xx + (ww / 3), yy + hh - 1);
    }

    if (area) {
	gdk_gc_set_clip_rectangle(fgc, NULL);
	gdk_gc_set_clip_rectangle(bgc, NULL);
    }

#if 0
    xthick = style->klass->xthickness;
    ythick = style->klass->ythickness;
#else
    xthick = 1;
    ythick = 1;
#endif

    gtk_paint_shadow(style, window, state_type, GTK_SHADOW_IN,
		     area, widget, detail,
		     x - xthick, y - ythick,
		     width + (xthick * 2), height + (ythick * 2));
}

static void
draw_option(GtkStyle * style,
	    GdkWindow * window,
	    GtkStateType state_type,
	    GtkShadowType shadow_type,
	    GdkRectangle * area,
	    GtkWidget * widget,
	    gchar * detail,
	    gint x,
	    gint y,
	    gint width,
	    gint height)
{
    GdkGC *fgc, *bgc, *dgc;

    g_return_if_fail (style != NULL);
    g_return_if_fail (window != NULL);

    if (width == -1 && height == -1)
	gdk_window_get_size(window, &width, &height);
    else if (width == -1)
	gdk_window_get_size(window, &width, NULL);
    else if (height == -1)
	gdk_window_get_size(window, NULL, &height);

    x -= 1;
    y -= 1;
    width += 2;
    height += 2;

    if (state_type == GTK_STATE_INSENSITIVE) {
	fgc = style->fg_gc[state_type];
	bgc = style->bg_gc[GTK_STATE_NORMAL];
    }
    else {
	fgc = style->black_gc;
	bgc = style->white_gc;
    }
    dgc  = whistler_dark_gc;

    if (area) {
	gdk_gc_set_clip_rectangle(fgc, area);
	gdk_gc_set_clip_rectangle(bgc, area);
	gdk_gc_set_clip_rectangle(dgc, area);
    }

    gdk_draw_arc(window, dgc, TRUE, x, y, width, height, 0, 360 * 64);
    gdk_draw_arc(window, dgc, FALSE, x, y, width, height, 0, 360 * 64);
    gdk_draw_arc(window, bgc, TRUE, x+1, y+1, width-2, height-2, 0, 360 * 64);
    gdk_draw_arc(window, bgc, FALSE, x+1, y+1, width-2, height-2, 0, 360 * 64);

    if (shadow_type == GTK_SHADOW_IN) {
	gdk_draw_arc(window, fgc, TRUE, x+4, y+4, width-8, height-8, 0, 360 * 64);
	gdk_draw_arc(window, fgc, FALSE, x+4, y+4, width-8, height-8, 0, 360 * 64);
    }

    if (area) {
	gdk_gc_set_clip_rectangle(fgc, NULL);
	gdk_gc_set_clip_rectangle(bgc, NULL);
	gdk_gc_set_clip_rectangle(dgc, NULL);
    }
}

static void
draw_cross(GtkStyle * style,
	   GdkWindow * window,
	   GtkStateType state_type,
	   GtkShadowType shadow_type,
	   GdkRectangle * area,
	   GtkWidget * widget,
	   gchar * detail,
	   gint x,
	   gint y,
	   gint width,
	   gint height)
{
    g_return_if_fail(style != NULL);
    g_return_if_fail(window != NULL);
}

static void
draw_ramp(GtkStyle * style,
	  GdkWindow * window,
	  GtkStateType state_type,
	  GtkShadowType shadow_type,
	  GdkRectangle * area,
	  GtkWidget * widget,
	  gchar * detail,
	  GtkArrowType arrow_type,
	  gint x,
	  gint y,
	  gint width,
	  gint height)
{
    g_return_if_fail(style != NULL);
    g_return_if_fail(window != NULL);
}

static void
draw_tab(GtkStyle * style,
	 GdkWindow * window,
	 GtkStateType state_type,
	 GtkShadowType shadow_type,
	 GdkRectangle * area,
	 GtkWidget * widget,
	 gchar * detail,
	 gint x,
	 gint y,
	 gint width,
	 gint height)
{
    g_return_if_fail(style != NULL);
    g_return_if_fail(window != NULL);

    gtk_paint_box(style, window, state_type, shadow_type, area, widget, detail,
		  x, y, width, height);
}

static void
draw_shadow_gap(GtkStyle * style,
		GdkWindow * window,
		GtkStateType state_type,
		GtkShadowType shadow_type,
		GdkRectangle * area,
		GtkWidget * widget,
		gchar * detail,
		gint x,
		gint y,
		gint width,
		gint height,
		GtkPositionType gap_side,
		gint gap_x,
		gint gap_width)
{
    GdkRectangle rect;

    g_return_if_fail(style != NULL);
    g_return_if_fail(window != NULL);

    if (width == -1 && height == -1)
	gdk_window_get_size(window, &width, &height);
    else if (width == -1)
	gdk_window_get_size(window, &width, NULL);
    else if (height == -1)
	gdk_window_get_size(window, NULL, &height);

    gtk_paint_shadow(style, window, state_type, shadow_type,
		     area, widget, detail, x, y, width, height);

    if (GTK_IS_NOTEBOOK(widget)) {
	gap_x += 1;
	gap_width -= 2;
    }

    switch (gap_side) {
    case GTK_POS_TOP:
	rect.x = x + gap_x;
	rect.y = y;
	rect.width = gap_width;
	rect.height = 1;
	break;

    case GTK_POS_BOTTOM:
	rect.x = x + gap_x;
	rect.y = y + height - 1;
	rect.width = gap_width;
	rect.height = 1;
	break;

    case GTK_POS_LEFT:
	rect.x = x;
	rect.y = y + gap_x;
	rect.width = 1;
	rect.height = gap_width;
	break;

    case GTK_POS_RIGHT:
	rect.x = x + width - 1;
	rect.y = y + gap_x;
	rect.width = 1;
	rect.height = gap_width;
	break;
    }

    gtk_style_apply_default_background(style, window,
				       widget && !GTK_WIDGET_NO_WINDOW(widget),
				       state_type, area,
				       rect.x, rect.y, rect.width, rect.height);
}

static void
draw_box_gap(GtkStyle * style,
	     GdkWindow * window,
	     GtkStateType state_type,
	     GtkShadowType shadow_type,
	     GdkRectangle * area,
	     GtkWidget * widget,
	     gchar * detail,
	     gint x,
	     gint y,
	     gint width,
	     gint height,
	     GtkPositionType gap_side,
	     gint gap_x,
	     gint gap_width)
{
    GdkRectangle rect;

    g_return_if_fail(style != NULL);
    g_return_if_fail(window != NULL);

    if (width == -1 && height == -1)
	gdk_window_get_size(window, &width, &height);
    else if (width == -1)
	gdk_window_get_size(window, &width, NULL);
    else if (height == -1)
	gdk_window_get_size(window, NULL, &height);

    gtk_paint_box(style, window, state_type, shadow_type, area, widget, detail,
		  x, y, width, height);

    if (GTK_IS_NOTEBOOK(widget)) {
	gap_x += 1;
	gap_width -= 2;
    }

    switch (gap_side) {
    case GTK_POS_TOP:
	rect.x = x + gap_x;
	rect.y = y;
	rect.width = gap_width;
	rect.height = 1;
	break;

    case GTK_POS_BOTTOM:
	rect.x = x + gap_x;
	rect.y = y + height - 1;
	rect.width = gap_width;
	rect.height = 1;
	break;

    case GTK_POS_LEFT:
	rect.x = x;
	rect.y = y + gap_x;
	rect.width = 1;
	rect.height = gap_width;
	break;

    case GTK_POS_RIGHT:
	rect.x = x + width - 1;
	rect.y = y + gap_x;
	rect.width = 1;
	rect.height = gap_width;
	break;
    }

    gtk_style_apply_default_background(style, window,
				       widget && !GTK_WIDGET_NO_WINDOW(widget),
				       state_type, area,
				       rect.x, rect.y, rect.width, rect.height);
}

static void
draw_extension(GtkStyle * style,
	       GdkWindow * window,
	       GtkStateType state_type,
	       GtkShadowType shadow_type,
	       GdkRectangle * area,
	       GtkWidget * widget,
	       gchar * detail,
	       gint x,
	       gint y,
	       gint width,
	       gint height,
	       GtkPositionType gap_side)
{
    gint xthick, ythick;
    GdkRectangle rect;

    g_return_if_fail(style != NULL);
    g_return_if_fail(window != NULL);

    if (width == -1 && height == -1)
	gdk_window_get_size(window, &width, &height);
    else if (width == -1)
	gdk_window_get_size(window, &width, NULL);
    else if (height == -1)
	gdk_window_get_size(window, NULL, &height);

    gtk_paint_box(style, window, state_type, shadow_type, area, widget, detail,
		  x, y, width, height);

#if 0
    xthick = style->klass->xthickness;
    ythick = style->klass->ythickness;
#else
    xthick = 1;
    ythick = 1;
#endif

    switch (gap_side) {
    case GTK_POS_TOP:
	rect.x = x + xthick;
	rect.y = y;
	rect.width = width - xthick * 2;
	rect.height = ythick;
	break;

    case GTK_POS_BOTTOM:
	rect.x = x + xthick;
	rect.y = y + height - ythick;
	rect.width = width - xthick * 2;
	rect.height = ythick;
	break;

    case GTK_POS_LEFT:
	rect.x = x;
	rect.y = y + ythick;
	rect.width = xthick;
	rect.height = height - ythick * 2;
	break;

    case GTK_POS_RIGHT:
	rect.x = x + width - xthick;
	rect.y = y + ythick;
	rect.width = xthick;
	rect.height = height - ythick * 2;
	break;
    }

    gtk_style_apply_default_background(style, window,
				       widget && !GTK_WIDGET_NO_WINDOW(widget),
				       state_type, area,
				       rect.x, rect.y, rect.width, rect.height);
}

static void
draw_focus(GtkStyle * style,
	   GdkWindow * window,
	   GdkRectangle * area,
	   GtkWidget * widget,
	   gchar * detail,
	   gint x,
	   gint y,
	   gint width,
	   gint height)
{
    GdkGC *gc;

    g_return_if_fail(style != NULL);
    g_return_if_fail(window != NULL);

    if (width == -1 && height == -1) {
	gdk_window_get_size(window, &width, &height);
	width -= 1;
	height -= 1;
    }
    else if (width == -1) {
	gdk_window_get_size(window, &width, NULL);
	width -= 1;
    }
    else if (height == -1) {
	gdk_window_get_size(window, NULL, &height);
	height -= 1;
    }

    gc = style->bg_gc[GTK_STATE_SELECTED];

    if (area) gdk_gc_set_clip_rectangle(gc, area);

    gdk_draw_rectangle(window, gc, FALSE, x, y, width, height);

    if (area) gdk_gc_set_clip_rectangle(gc, NULL);
}

static void
draw_slider(GtkStyle * style,
	    GdkWindow * window,
	    GtkStateType state_type,
	    GtkShadowType shadow_type,
	    GdkRectangle * area,
	    GtkWidget * widget,
	    gchar * detail,
	    gint x,
	    gint y,
	    gint width,
	    gint height,
	    GtkOrientation orientation)
{
    gint xthick, ythick;

    g_return_if_fail(style != NULL);
    g_return_if_fail(window != NULL);

    if (width == -1 && height == -1)
	gdk_window_get_size(window, &width, &height);
    else if (width == -1)
	gdk_window_get_size(window, &width, NULL);
    else if (height == -1)
	gdk_window_get_size(window, NULL, &height);

    if (state_type == GTK_STATE_INSENSITIVE)
	gtk_paint_box(style, window, state_type, GTK_SHADOW_OUT, area, widget,
		      detail, x, y, width, height);
    else
	gtk_paint_box(style, window, state_type, shadow_type, area, widget,
		      "button", x, y, width, height);

#if 0
    xthick = style->klass->xthickness;
    ythick = style->klass->ythickness;
#else
    xthick = 1;
    ythick = 1;
#endif

    if (orientation == GTK_ORIENTATION_HORIZONTAL)
	gtk_paint_vline(style, window, state_type, area, widget, detail,
			ythick, height - ythick - 1, width / 2);
    else
	gtk_paint_hline(style, window, state_type, area, widget, detail,
			xthick, width - xthick - 1, height / 2);
}

static void
draw_handle(GtkStyle * style,
	    GdkWindow * window,
	    GtkStateType state_type,
	    GtkShadowType shadow_type,
	    GdkRectangle * area,
	    GtkWidget * widget,
	    gchar * detail,
	    gint x,
	    gint y,
	    gint width,
	    gint height,
	    GtkOrientation orientation)
{
    gint xthick, ythick;
    gint xx, yy;
    GdkGC *lightgc, *darkgc;
    GdkRectangle rect;
    GdkRectangle dest;
    gboolean intersect;

    g_return_if_fail (style != NULL);
    g_return_if_fail (window != NULL);

    if (width == -1 && height == -1)
	gdk_window_get_size(window, &width, &height);
    else if (width == -1)
	gdk_window_get_size(window, &width, NULL);
    else if (height == -1)
	gdk_window_get_size(window, NULL, &height);

    gtk_paint_box(style, window, state_type, shadow_type, area, widget,
	          detail, x, y, width, height);

    lightgc = style->base_gc[GTK_STATE_NORMAL];
    darkgc  = whistler_dark_gc;

#if 0
    xthick = style->klass->xthickness;
    ythick = style->klass->ythickness;
#else
    xthick = 1;
    ythick = 1;
#endif

    rect.x = x + xthick;
    rect.y = y + ythick;
    rect.width = width - (xthick * 2);
    rect.height = height - (ythick * 2);

    if (area)
	intersect = gdk_rectangle_intersect(area, &rect, &dest);
    else {
	intersect = TRUE;
	dest = rect;
    }

    if (!intersect)
	return;

#define DRAW_POINT(w, gc, clip, xx, yy)			\
    {							\
	if ((xx) >= (clip).x				\
	    && (yy) >= (clip).y				\
	    && (xx) < (clip).x + (clip).width		\
	    && (yy) < (clip).y + (clip).height)		\
		gdk_draw_point ((w), (gc), (xx), (yy));	\
    }

    if (area) {
	gdk_gc_set_clip_rectangle(lightgc, area);
	gdk_gc_set_clip_rectangle(darkgc,  area);
    }

    for (yy = y + ythick; yy < (y + height - ythick); yy += 3) {
	for (xx = x + xthick; xx < (x + width - xthick); xx += 6) {
	    DRAW_POINT (window, lightgc, dest, xx, yy);
	    DRAW_POINT (window, darkgc,  dest, xx + 1, yy + 1);

	    DRAW_POINT (window, lightgc, dest, xx + 3, yy + 1);
	    DRAW_POINT (window, darkgc,  dest, xx + 4, yy + 2);
	}
    }

    if (area) {
	gdk_gc_set_clip_rectangle(lightgc, NULL);
	gdk_gc_set_clip_rectangle(darkgc,  NULL);
    }
}
